(module tl-tests mzscheme
  (require "test.scm")
  
  
  ;; we just avoid testing mzscheme for these
  (define (to-mz x) 1)
  
  (define tl-ts
    (test-suite
     "tl.scm"
     to-mz
     equal?

     (test "Basic beginD 1"
           '(store () (beginD (+ 1 2) (+ 3 4)))
           '(store () 7))
     
     (test "Basic beginD 2"
           '(store () (+ 1 (begin (+ 1 2) (+ 3 4))))
           '(store () 8))
     
     (test "Basic define 1"
           '(store () (define x 1) (set! x 2))
           '(store ((x 2)) unspecified))
     
     (test "Basic define 2"
           '(store () (define x 1) (define y x))
           '(store ((x 1) (y 1)) unspecified))
     
     (test "Basic define 3"
           '(store () (define x 1) (define y (+ x x)))
           '(store ((x 1) (y 2)) unspecified))
     
     (test "Basic fn app 1"
           '(store () ((lambda (x) x) 1))
           '(store () 1))
     
     (test "BeginD/define 1"
           '(store () (beginD (define x 1) (define y x)))
           '(store ((x 1) (y 1)) unspecified))
     
     (test "BeginD/define 2"
           '(store () (beginD (define x 1) (define y 2)) (beginD (define z (+ x y))))
           '(store ((x 1) (y 2) (z 3)) unspecified))
     
     (test "top-level call/cc 1"
           '(store () (define k (call/cc (lambda (x) x))) (define y (k 1)))
           '(store ((k 1)) unspecified))
     
     (test "top-level call/cc 2" 
           '(store ()
                   (beginD (define x 1)
                           (define k (call/cc (lambda (x) x)))
                           (define y 1))
                   (set! x 2)
                   (set! y 2)
                   (k 1))
           '(store ((x 2)
                    (k 1)
                    (y 2))
                   unspecified))
     
     (test "call/cc treated as a value"
           '(store () (define x (((lambda (x) call/cc) call/cc) (lambda (k) (k 1)))))
           '(store ((x 1)) unspecified))
     
     (test "begin doesn't become beginD"
           '(store ()
                   (define x 1)
                   (define k 2)
                   (define f (lambda () (begin (call/cc (lambda (k2) (set! k k2))) (set! x (+ x 1)))))
                   (f)
                   (set! x 3)
                   (k 1)
                   (set! k 11)
                   (set! f 12))
           '(store ((x 4)
                    (k 11)
                    (f 12))
                   unspecified))
     
     ;; test recursive definitions
     (test "self-referential defs 1"
           '(store ()
                   (define f
                     (lambda (x y)
                       (x (lambda () (f y x)))))
                   (f (lambda (fc) (fc))
                      (lambda (fc) 1)))
           '(store ((f
                     (lambda (x y)
                       (x (lambda () (f y x))))))
                   1))
     
     (test "self-referential defs 2"
           '(store () (define x 1) (define x (+ x x)))
           '(store ((x 2)) unspecified))
     
     (test "self-referential defs 3"
           '(store () x)
           '(store () "reference to free identifier: x"))
     
     (test "self-referential defs 4"
           '(store () (define x x))
           '(store () "reference to free identifier: x"))
     
     
     (test "self-referential defs 5"
           '(store () (define f (+ (+ 1 2) f)))
           '(store () "reference to free identifier: f"))
     
     (test "self-referential defs 6" 
           '(store () (define f (+ (+ 1 2) (set! f 2))))
           '(store () "attempt to set! free identifier: f"))

     (test "begin w/set! following continuation grab"
           '(store ()
                   (define count 0)
                   (beginD
                     (define k (call/cc (lambda (x) x)))
                     (set! count (+ count 1)))
                   (k (lambda (x) x)))
           '(store ((count 1)
                    (k (lambda (x) x)))
                   unspecified))
     
     
     (test "grab cont with two expressions after and one before"
           '(store ()
                   (define count 0)
                   (beginD
                     (define x 1)
                     (define k (call/cc (lambda (x) x)))
                     (set! count (+ count 1))
                     (define y 1))
                   (set! x (+ x 1))
                   (set! y (+ y 1))
                   (k (lambda (x) x)))
           '(store ((count 1)
                    (x 2)
                    (k (lambda (x) x))
                    (y 2))
                   unspecified))
     
     (test "inner define becomes top-level"
           '(store 
             ()
             (define x 1)
             (define k 0)
             ((lambda (ignore) 
                (begin ((call/cc (lambda (k1) 
                                   (begin
                                     (set! k k1)
                                     (lambda (x) x))))
                        1)
                       (set! x (+ x 1))))
              0)
             (k (lambda (x) (call/cc (lambda (k2)
                                       (begin (set! k k2)
                                              (lambda (x) x))))))
             (k (lambda (x) x))
             (set! k 11)
             x)
           '(store 
             ((x 4)
              (k 11))
             4))
     ))
  
  (provide tl-ts))
